//
// File:        CxxStubSource.java
// Package:     gov.llnl.babel.backend.ucxx
// Revision:    @(#) $Revision: 5423 $
// Date:        $Date: 2006-04-20 15:14:51 -0700 (Thu, 20 Apr 2006) $
// Description: Write Cxx extension header file for a BABEL extendable
// 
// This is typically directed by GenCxxClient.
// Copyright (c) 2000-2001, The Regents of the University of Calfornia.
// Produced at the Lawrence Livermore National Laboratory.
// Written by the Components Team <components@llnl.gov>
// UCRL-CODE-2002-054
// All rights reserved.
// 
// This file is part of Babel. For more information, see
// http://www.llnl.gov/CASC/components/. Please read the COPYRIGHT file
// for Our Notice and the LICENSE file for the GNU Lesser General Public
// License.
// 
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License (as published by
// the Free Software Foundation) version 2.1 dated February 1999.
// 
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
// conditions of the GNU Lesser General Public License for more details.
// 
// You should have recieved a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software Foundation,
// Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

package gov.llnl.babel.backend.ucxx;

import gov.llnl.babel.BabelConfiguration;
import gov.llnl.babel.backend.CodeGenerationException;
import gov.llnl.babel.backend.ucxx.Cxx;
import gov.llnl.babel.backend.rmi.RMIStubSource;
import gov.llnl.babel.backend.rmi.RMI;
import gov.llnl.babel.backend.IOR;
import gov.llnl.babel.backend.LevelComparator;
import gov.llnl.babel.backend.writers.LanguageWriterForCxx;
import gov.llnl.babel.symbols.Comment;
import gov.llnl.babel.symbols.Extendable;
import gov.llnl.babel.symbols.Class;
import gov.llnl.babel.symbols.Method;
import gov.llnl.babel.symbols.SymbolID;
import gov.llnl.babel.symbols.SymbolTable;
import gov.llnl.babel.symbols.Type;
import java.util.Arrays;
import java.util.Iterator;
import java.util.Collection;

/**
 * Create and write a header for a Cxx C extension class to wrap a 
 * BABEL extendable in a Cxx object. The header has to expose a 
 * function to create a wrapped IOR, a function to check if a 
 * <code>PyObject</code> is an instance of this extension type, and
 * an import macro.
 */
public class CxxStubSource {
  private Extendable d_ext = null;
 private LanguageWriterForCxx d_writer = null;
  private String d_self = null;
  
  /**
   * Create an object capable of generating the header file for a
   * BABEL extendable.
   *
   * @param ext   an interface or class symbol that needs a header
   *              file for a Cxx C extension class.
   */
  public CxxStubSource(Extendable ext) {
    d_ext = ext;
  }
  /**
   * Create an object capable of generating the header file for a
   * BABEL extendable.
   *
   * @param ext   an interface or class symbol that needs a header
   *              file for a Cxx C extension class.
   */
  public CxxStubSource(Extendable ext,
                       LanguageWriterForCxx writer) {
    d_ext = ext;
    d_writer = writer;
  }
  
  /**
   * This is a convenience utility function specifically for the generation
   * of super "Stub" functions in the Impl files. 
   * The output stream is not closed on exit.  A code
   * generation exception is thrown if an error is detected.
   *
   * @param cls The class in which these supers are to be generated  
   * 
   * @param writer the output writer to which the stub source will
   *               be written. This will not be closed.
   *    
   * @exception gov.llnl.babel.backend.CodeGenerationException
   *    this is a catch all exception. It can be caused by I/O trouble or
   *    violations of the data type invariants.
   */
  public static void generateSupers(Class             cls,
                                    LanguageWriterForCxx writer)
    throws CodeGenerationException
  {
    CxxStubSource source = new CxxStubSource(cls, writer);
    source.generateSupers();
  }  

    /**
   * This special function is only to be used when generating super
   * functions for the Super class in an Impl.  It may only be called though
   * the static method above, "generateSupers"
   *
   */
  private void generateSupers() 
    throws CodeGenerationException {
    Class cls = (Class) d_ext;
    SymbolID clsID = cls.getSymbolID();
    Collection methods = cls.getOverwrittenClassMethods();
    for(Iterator mit = methods.iterator(); mit.hasNext();){
      Method method = (Method)mit.next();
      generateMethodDispatch( method, "user defined static method", true, true);
      d_writer.println();
      if(method.hasRarray()) {
        generateMethodDispatch( method, "user defined static method",true, false);
        d_writer.println();
      }
    }
  }

  /**
   * Generate the header file for the extendable with which this object was
   * created.
   *
   * @exception gov.llnl.babel.backend.CodeGenerationException
   *    this is a catch all exception for problems during the code
   *    generation phase.
   */
  public synchronized void generateCode() throws CodeGenerationException {
    if ( d_ext.isInterface() ) { 
      d_self = "loc_self->d_object";
    } else { 
      d_self = "loc_self";
    }

    try { 
      d_writer = Cxx.createSource( d_ext, Cxx.FILE_ROLE_STUB, "STUBSRCS" );
      d_writer.println();

      writeIncludes();

      //No Language specific RMI initialization for C++
      d_writer.println("#define "+RMI.LangSpecificInit());

      d_writer.writeComment("connect_loaded is a boolean value showing if "+
                            "the IHConnect for this object has been loaded "+
                            "into the connectRegistry",false);
      d_writer.println("static int connect_loaded = 0;");
      
      d_writer.println("extern \"C\" {");
      d_writer.tab();
      RMIStubSource.generateCode(d_ext,d_writer);
      d_writer.backTab();
      d_writer.println("}");

      writeUserDefinedMethods();

      writeConstructors();

      writeCastingOperators();

      if ( !d_ext.isInterface()) {
        writeDynamicImplStuff();
      }
      /*//TODO: this is temporary 'til we figure out what to do about connectI
      
      SymbolID id = d_ext.getSymbolID();
      String ior_t = IOR.getObjectName(id);
      String ior_name = IOR.getSymbolName(id);
      d_writer.writeCommentLine("redeclared here for other UCxx stub .cc's");
      d_writer.println("extern \"C\" {");
      d_writer.tab();
      d_writer.println(ior_t + "* " + ior_name + "__connectI( const char* url, sidl_bool ar, sidl_BaseInterface__object**_ex) {");
      d_writer.tab();
      d_writer.println("return " + ior_name + "__remoteConnect( url, ar, _ex);");  
      d_writer.backTab();
      d_writer.println("}");
      d_writer.backTab();
      d_writer.println("}");
      d_writer.println();
      */
    } finally { 
      if (d_writer != null) {
        d_writer.close();
        d_writer = null;
      }
    }
  }

  private void writeIncludes() throws CodeGenerationException { 
    SymbolID id = d_ext.getSymbolID();
    d_writer.generateInclude( Cxx.generateFilename( d_ext.getSymbolID(), 
                                                    Cxx.FILE_ROLE_STUB, 
                                                    Cxx.FILE_TYPE_CXX_HEADER ),
                              true);
    d_writer.generateInclude( Cxx.generateFilename( "sidl.BaseInterface",
                                                    Cxx.FILE_ROLE_STUB, 
                                                    Cxx.FILE_TYPE_CXX_HEADER
                                                    ),
                              true );
    d_writer.generateInclude(  Cxx.generateFilename( "sidl.BaseClass",
                                                    Cxx.FILE_ROLE_STUB, 
                                                    Cxx.FILE_TYPE_CXX_HEADER
                                                    ),
                               true );
    d_writer.generateInclude( Cxx.generateFilename( "sidl.RuntimeException",
                                                    Cxx.FILE_ROLE_STUB, 
                                                    Cxx.FILE_TYPE_CXX_HEADER), 
                              true );
    if ( d_ext.hasExceptionThrowingMethod(true) ) { 
      d_writer.generateInclude( Cxx.generateFilename( "sidl.BaseException",
                                                    Cxx.FILE_ROLE_STUB, 
                                                    Cxx.FILE_TYPE_CXX_HEADER), 
                                true );
    }

   
    d_writer.generateInclude(  Cxx.generateFilename( "sidl.rmi.Call",
                                                    Cxx.FILE_ROLE_STUB, 
                                                    Cxx.FILE_TYPE_CXX_HEADER
                                                    ),
                               true );
    d_writer.generateInclude(  Cxx.generateFilename( "sidl.rmi.Return",
                                                    Cxx.FILE_ROLE_STUB, 
                                                    Cxx.FILE_TYPE_CXX_HEADER
                                                    ),
                               true );
    d_writer.generateInclude(  Cxx.generateFilename( "sidl.rmi.InstanceHandle",
                                                    Cxx.FILE_ROLE_STUB, 
                                                    Cxx.FILE_TYPE_CXX_HEADER
                                                    ),
                               true );
    
    d_writer.generateInclude( "sidl_String.h", false );
    d_writer.generateInclude("sidl_rmi_ConnectRegistry.h", false);
    if (!BabelConfiguration.isSIDLBaseClass(id)) {
      d_writer.generateInclude("babel_config.h", false);
      d_writer.printlnUnformatted("#ifdef SIDL_DYNAMIC_LIBRARY");
      d_writer.printlnUnformatted("#include <stdio.h>");
      d_writer.printlnUnformatted("#include <stdlib.h>");
      d_writer.generateInclude("sidl_Loader.hxx", false);
      d_writer.generateInclude("sidl_DLL.hxx", false);

      d_writer.printlnUnformatted("#endif");
    }
    
    //Generate method dependency includes
    Cxx.generateSourceIncludes(d_writer, d_ext);
    d_writer.println();
  } 

  private void writeConstructors() { 
    SymbolID id = d_ext.getSymbolID();
    final String fullName = Cxx.getObjectName(id);
    final String fullNameWithoutLeadingColons = 
                                  Cxx.getSymbolNameWithoutLeadingColons(id,"");
    final String name = id.getShortName();
    final String ior_name = IOR.getSymbolName(id);

    if( !d_ext.isAbstract() ) { 
      d_writer.writeCommentLine("static constructor");
      d_writer.println(fullName);
      d_writer.println(fullNameWithoutLeadingColons + "::_create() {");
      d_writer.tab();
      d_writer.println(IOR.getExceptionFundamentalType() + "throwaway_exception;");
      d_writer.println( Cxx.getObjectName(id) 
        + " self( (*_get_ext()->createObject)(&throwaway_exception), false );");
      d_writer.println("return self;");
      d_writer.backTab();

      d_writer.println("}");
      d_writer.println();

      d_writer.writeCommentLine("remote constructor");
      d_writer.println(fullName);
      d_writer.println(fullNameWithoutLeadingColons + "::_create(" + 
		       "const std::string& url) {");
      d_writer.tab();
      d_writer.println( "ior_t* ior_self;");
      d_writer.println( "sidl_BaseInterface__object* _ex = 0;");
      d_writer.println( "ior_self = " + ior_name + "__remoteCreate( url.c_str(), &_ex );" );
      d_writer.println( "if (_ex != 0 ) {");
      d_writer.tab();
      d_writer.println( "; //TODO: handle exception" );
      d_writer.backTab();
      d_writer.println("}");
      d_writer.println( "return " + fullName + "( ior_self, false );" );
      d_writer.backTab();
      d_writer.println("}");
      d_writer.println();
    }

    d_writer.writeCommentLine("remote connector 2");
    d_writer.println(fullName);
    d_writer.println(fullNameWithoutLeadingColons + "::_connect(" + 
		     "const std::string& url, const bool ar ) {");
    d_writer.tab();
    d_writer.println( "ior_t* ior_self;");
    d_writer.println( "sidl_BaseInterface__object* _ex = 0;");
    d_writer.println( "ior_self = " + ior_name + "__remoteConnect( url.c_str(), ar?TRUE:FALSE, &_ex );" );
    d_writer.println( "if (_ex != 0 ) {");
    d_writer.tab();
    d_writer.println( "; //TODO: handle exception" );
    d_writer.backTab();
    d_writer.println("}");
    d_writer.println( "return " + fullName + "( ior_self, false );" );
    d_writer.backTab();
    d_writer.println("}");
    d_writer.println();
      
    d_writer.writeCommentLine("copy constructor");
    d_writer.println( fullNameWithoutLeadingColons + "::" + name 
      + " ( const " + fullName + "& original ) {");
    d_writer.tab();
    //Cast changed here
    d_writer.println("d_self = "+Cxx.getObjectName(id)+"::_cast(original._get_ior());");
    d_writer.println("d_weak_reference = false;");
    d_writer.backTab();
    d_writer.println("}");
    d_writer.println();
    
    d_writer.writeCommentLine("assignment operator");
    d_writer.println( fullName + "&");
    d_writer.println( fullNameWithoutLeadingColons + "::operator=( const " 
      + fullName + "& rhs ) {");
    d_writer.tab();
    d_writer.println("if ( d_self != rhs.d_self ) {");
    d_writer.tab();
    d_writer.println("if ( d_self != 0 ) {");
    d_writer.tab();
    d_writer.println("deleteRef();");
    d_writer.backTab();
    d_writer.println("}");
    d_writer.println("d_self = "+Cxx.getObjectName(id)+"::_cast(rhs._get_ior());"); 
    d_writer.writeCommentLine("note _cast incremements the reference count");
    d_writer.println("d_weak_reference = false;");
    d_writer.backTab();
    d_writer.println("}");
    d_writer.println("return *this;");
    d_writer.backTab();
    d_writer.println("}");
    d_writer.println();
  }

  private void writeCastingOperators() throws CodeGenerationException 
  {  
    SymbolID id = d_ext.getSymbolID();
    final String fullName = Cxx.getObjectName(id);
    final String fullNameWithoutLeadingColons = 
                                   Cxx.getSymbolNameWithoutLeadingColons(id,"");
    final String name = id.getShortName();   
    final String ior_name = IOR.getSymbolName(id);
    
    d_writer.writeCommentLine("conversion from ior to C++ class");
    d_writer.println( fullNameWithoutLeadingColons + "::" + name 
      + " ( " + fullName + "::ior_t* ior ) ");
    
    d_writer.println("   : StubBase("+Cxx.reinterpretCast("void*","ior")
      + ") { }");
    d_writer.println();

    d_writer.writeCommentLine("Alternate constructor: does not call addRef()");
    d_writer.writeCommentLine("(sets d_weak_reference=isWeak)");
    d_writer.writeCommentLine("For internal use by Impls (fixes bug#275)");
    d_writer.println( fullNameWithoutLeadingColons + "::" + name +" ( " 
      + fullName + "::ior_t* ior, bool isWeak ) : ");
    d_writer.println( "StubBase(" + Cxx.reinterpretCast("void*","ior") 
      + ", isWeak){ ");
    d_writer.println("}");
    d_writer.println();
    
    d_writer.writeCommentLine("exec has special argument passing to avoid #include circularities");
    d_writer.println("void " + fullName + "::_exec( const std::string& methodName, ");
    d_writer.println("                        UCXX_LOCAL sidl::rmi::Call& inArgs,");
    d_writer.println("                        UCXX_LOCAL sidl::rmi::Return& outArgs) { ");

    d_writer.tab();
    d_writer.println(fullName + "::ior_t* loc_self = _get_ior();");     
                     // + Cxx.reinterpretCast("ior_t*","this->d_self") + ";");
    d_writer.println(IOR.getExceptionFundamentalType() + "throwaway_exception;");
    d_writer.println("(*loc_self->d_epv->f__exec)(" + d_self + ",");
    d_writer.println("                              methodName.c_str(),");
    d_writer.println("                              inArgs._get_ior(),");
    d_writer.println("                              outArgs._get_ior(),");
    d_writer.println("                              &throwaway_exception);");
    d_writer.backTab();
    d_writer.println("}");
    d_writer.println();
      
    generateMethodDispatch(IOR.getBuiltinMethod(IOR.GETURL, d_ext.getSymbolID(), false),
                           "get URL of the implementation of this object", false,false);

    d_writer.writeCommentLine("protected method that implements casting");
    d_writer.println(IOR.getObjectName(id)+ "* " + fullNameWithoutLeadingColons 
                     + "::_cast(const void* src)"); 
    d_writer.println("{");

    d_writer.tab();

    d_writer.println("ior_t* cast = NULL;"); 
 
    d_writer.println();
    d_writer.println("if(!connect_loaded) {");
    d_writer.tab();
    d_writer.println(IOR.getExceptionFundamentalType() + "throwaway_exception;");
    
    d_writer.println("sidl_rmi_ConnectRegistry_registerConnect(\""+id.getFullName()+
                     "\", (void*)"+IOR.getSymbolName(id)+"__IHConnect, &throwaway_exception);");
    d_writer.println("connect_loaded = 1;");
    d_writer.backTab();
    d_writer.println("}");

    d_writer.println("if ( src != 0 ) {");
    d_writer.tab();
    d_writer.writeCommentLine("Actually, this thing is still const");
    d_writer.println("void* tmp = const_cast<void*>(src);"); 
    d_writer.println(IOR.getExceptionFundamentalType() + "throwaway_exception;");
    d_writer.println("sidl_BaseInterface base = " +  
                     Cxx.reinterpretCast("const sidl_BaseInterface", "tmp") + ";");
    d_writer.print("cast = "); 
    d_writer.print(Cxx.reinterpretCast( "ior_t*",
                                        "(*base->d_epv->f__cast)(base->d_object, \""+
                                        id.getFullName()+"\", &throwaway_exception)"));
    
    d_writer.println(";");
    d_writer.backTab();
    d_writer.println("}");    
    d_writer.println("return cast;");//+fullNameWithoutLeadingColons+"(cast);");
    d_writer.backTab();
    d_writer.println("}");
    d_writer.println();
    
  }

  private void writeDynamicImplStuff() throws CodeGenerationException { 
    SymbolID id = d_ext.getSymbolID();
    final String extName = IOR.getExternalName(id);
    final String fullName = Cxx.getObjectName(id);
    final String fullNameWithoutLeadingColons = 
                                  Cxx.getSymbolNameWithoutLeadingColons(id,"");
    d_writer.writeCommentLine("Static data type");
    d_writer.println("const " + fullName + "::ext_t * " 
      + fullNameWithoutLeadingColons + "::s_ext = 0;");
    d_writer.println();

    d_writer.writeCommentLine("private static method to get static data type");
    d_writer.println("const " + fullName + "::ext_t *");
    d_writer.println( fullNameWithoutLeadingColons + "::_get_ext()");
    d_writer.println("  throw ( " + Cxx.prependGlobalUCxx() 
      + "::sidl::NullIORException)");
    d_writer.println("{");
    d_writer.tab();
    d_writer.println("if (! s_ext ) {");
    d_writer.tab();
    if (BabelConfiguration.isSIDLBaseClass(id)) { 
      d_writer.println("s_ext = " + IOR.getExternalFunc(id) + "();");
    } else {
      d_writer.printlnUnformatted("#ifdef SIDL_STATIC_LIBRARY");
      d_writer.println("s_ext = " + IOR.getExternalFunc(id) + "();");
      d_writer.printlnUnformatted("#else");
      d_writer.println("s_ext = ("+extName+"*)sidl_dynamicLoadIOR(\""+id.getFullName()+"\",\""+IOR.getExternalFunc(id)+"\") ;");
      d_writer.printlnUnformatted("#endif");
    }
    d_writer.backTab();
    d_writer.println("}");
    d_writer.println("return s_ext;");
    d_writer.backTab();
    d_writer.println("}");   
    d_writer.println();    
  }

  private void writeUserDefinedMethods() throws CodeGenerationException { 
    d_writer.beginBoldComment();
    d_writer.println("User Defined Methods");
    d_writer.endBoldComment();

    Iterator m = null;
    m = Cxx.getStubMethodList(d_ext).iterator();
    while (m.hasNext()) {
      final Method method = (Method) m.next();
      final boolean isStatic = method.isStatic();
      final String desc = isStatic ? "user defined static method" :
        "user defined non-static method.";
      if(!Cxx.inlineStub(method)) {
        generateMethodDispatch( method, desc, false, true);
        d_writer.println();
        if ((!isStatic) && 
            (method.getCommunicationModifier() == Method.NONBLOCKING) ) { 
          Method send = method.spawnNonblockingSend();
          generateMethodDispatch(send, desc, false, true);
          d_writer.println();
          Method recv = method.spawnNonblockingRecv();
          generateMethodDispatch(recv, desc, false, true);
          d_writer.println();
        }
        if(method.hasRarray()) {
          generateMethodDispatch( method, desc, false, false);
          d_writer.println();
          if ((!isStatic) && 
              (method.getCommunicationModifier() == Method.NONBLOCKING) ) { 
            Method send = method.spawnNonblockingSend();
            generateMethodDispatch(send, desc, false, false);
            d_writer.println();
            Method recv = method.spawnNonblockingRecv();
            generateMethodDispatch(recv, desc, false, false);
            d_writer.println();
          }
        }
      }
    }

    d_writer.beginBoldComment();
    d_writer.println("End User Defined Methods");
    d_writer.println("(everything else in this file is specific to");
    d_writer.println(" Babel's C++ bindings)");
    d_writer.endBoldComment();
  }

  /** 
   *  Function generates function definitions. 
   */
  private void generateMethodDispatch( Method m, String altcomment, 
                                       boolean isSuper, boolean rarrays ) 
    throws CodeGenerationException 
  { 
    if ( m == null ) { return; }

    String shortMethodName = m.getShortMethodName();
    String className = d_ext.getSymbolID().getShortName();
    Comment comment = m.getComment();
    Type return_type = m.getReturnType();

    d_writer.writeComment( comment, altcomment );
    if ( shortMethodName.equals(className) ) { 
      shortMethodName = "f_" + shortMethodName;
      System.out.println("WARNING: gov.llnl.babel.backend.UCxx.CxxStubSource: "
        + "sidl / C++ conflict!");
      System.out.println("         methodName == className is not allowed in "
        + "C++");
      System.out.println("         (this is restricted to constructors in C++"
        + ")");
      System.out.println("         changing to " + className + "::" 
        + shortMethodName + "()");
    }

    //Output function deleration
    d_writer.println(Cxx.generateFunctionDeclaration(m, d_ext, d_self, 
                                                     isSuper, rarrays));

    d_writer.println();
    d_writer.println("{");

    d_writer.tab();
    if ( (! m.isStatic()) &&  BabelConfiguration.getInstance().makeCxxCheckNullIOR()) { 

      d_writer.println("if ( d_self == 0 ) {");
      d_writer.tab();
      d_writer.println("throw " + Cxx.prependGlobalUCxx() 
        + "::sidl::NullIORException( ::std::string (");
      d_writer.tab();
      d_writer.disableLineBreak();
      d_writer.println( "\"" + "Null IOR Pointer in \\\"" 
                        + Cxx.getMethodStubName(d_ext.getSymbolID(),shortMethodName, isSuper) 
        + "()\\\"\"");
      d_writer.enableLineBreak();
      d_writer.backTab();
      d_writer.println("));");
      d_writer.backTab();
      d_writer.println("}");
    }

    //Output almost nothing!
    d_writer.println(Cxx.generateInitialization(m, d_ext, d_self));

    if ( ! m.isStatic() ) { 
      if (shortMethodName.equals("addRef") ||
          shortMethodName.equals("deleteRef") ) { 
        d_writer.println("if ( !d_weak_reference ) {");
        d_writer.tab();
      }
    }

    //Prepare for IOR call!
    d_writer.println(Cxx.generatePreIORCall(m, d_ext, d_self, isSuper, rarrays));


    if(isSuper) {
      Class cls = (Class) d_ext;
      String tmp_self = Cxx.reinterpretCast(IOR.getObjectName(cls.getParentClass().getSymbolID())
                                            + "*", "superSelf");
      d_writer.println(Cxx.generateIORCall(m, d_ext, tmp_self, isSuper, rarrays));
    } else {
      d_writer.println(Cxx.generateIORCall(m, d_ext, d_self, isSuper, rarrays));
    }

    if (m.getThrows().size()>0) { // if throws exception
      d_writer.println("if (_exception != 0 ) {");      
      d_writer.tab();
      d_writer.println("void * _p = 0;");
      d_writer.println(IOR.getExceptionFundamentalType() + "throwaway_exception;");
      Object [] exceptions = m.getThrows().toArray();
      Arrays.sort(exceptions, new LevelComparator(SymbolTable.getInstance()));
      
      for( int i=0; i<exceptions.length; ++i) {
        SymbolID exid = (SymbolID)exceptions[i];
        d_writer.println("if ( (_p=(*(_exception->d_epv->f__cast))(_exception->d_object,"
          + " \"" + exid.getFullName() + "\", &throwaway_exception)) != 0 ) {");
        d_writer.tab();
        d_writer.println(IOR.getObjectName(exid) + " * _realtype = " 
          + Cxx.reinterpretCast(IOR.getObjectName(exid)+"*" , "_p") + ";");
        d_writer.println("(*_exception->d_epv->f_deleteRef)(_exception->d_object, &throwaway_exception);");
        d_writer.writeCommentLine("Note: alternate constructor does not "
          + "increment refcount.");
        d_writer.println("throw " + Cxx.getObjectName(exid) 
          + "( _realtype, false );");
        d_writer.backTab();
        d_writer.println("}");
      }
      d_writer.backTab();
      d_writer.println("}");
    }

    //Clean up from return from IOR
    d_writer.println(Cxx.generatePostIORCall(m, d_ext, d_self, isSuper, rarrays));

    //d_writer.println( post_ior.toString().trim() );
    if ( ! m.isStatic() ) { 
      if ( shortMethodName.equals("addRef") 
         || shortMethodName.equals("deleteRef") ) { 
        if ( shortMethodName.equals("deleteRef") ) { 
          //d_writer.println("d_self = 0;");
          d_writer.println("_set_ior(0);");
        }
        d_writer.backTab();
        d_writer.println("}");
      }
    }

    if ( return_type.getType() != Type.VOID ) { 
      d_writer.println("return _result;");
    }
    d_writer.backTab();
    d_writer.println("}");
    d_writer.println();
  }
}
